class RegtaException(Exception):
    pass


class StopService(RegtaException):
    pass


class IncorrectJobType(RegtaException):
    def __init__(self, job, scheduler):
        message = f"{job.__class__.__name__} is incorrect job type for {scheduler.__class__.__name__}"
        super().__init__(message)



from typing import Dict, Union

from logging import Formatter, INFO, Logger, LoggerAdapter, StreamHandler
import sys
import traceback

from click import style
from click.utils import auto_wrap_for_ansi, resolve_color_default, should_strip_ansi, strip_ansi, WIN

prod_log_format = '%(asctime)s [%(job)s] [%(levelname)s] - %(message)s'
empty_log_format = '%(message)s'


class BorgSingletonMeta(type):
    _instances: Dict[type, object] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class ClickFormatter(Formatter):
    def __init__(self, *args, use_ansi: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_ansi = use_ansi

    def formatException(self, ei) -> str:
        exc_class, exc_info, exc_traceback = ei

        chain = []
        for line in traceback.format_tb(exc_traceback):
            for split_line in line.split('\n'):
                if not split_line:
                    continue
                chain.append(split_line)

        exception_str = f'{exc_class.__name__}: {str(exc_info)}'

        return (
            '\n'.join(map(lambda line: (style(line, fg='yellow') if self.use_ansi else line), chain))
            + '\n'
            + (style(exception_str, fg='red') if self.use_ansi else exception_str)
        )


class ClickStreamHandler(StreamHandler, metaclass=BorgSingletonMeta):
    def __init__(self, use_ansi: bool):
        super().__init__(stream=sys.stdout)
        self.use_ansi = use_ansi

    def emit(self, record):
        try:
            out = self.format(record) + self.terminator
            stream = self.stream

            color = resolve_color_default(color=self.use_ansi)
            if should_strip_ansi(stream, color):
                out = strip_ansi(out)
            elif WIN:
                if auto_wrap_for_ansi is not None:
                    stream = auto_wrap_for_ansi(stream)  # type: ignore
                elif not color:
                    out = strip_ansi(out)

            stream.write(out)
            self.flush()
        except RecursionError:
            raise
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)


class DefaultLogger(Logger, metaclass=BorgSingletonMeta):
    pass


def make_default_logger(use_ansi: bool, fmt: str = prod_log_format) -> DefaultLogger:
    formatter = ClickFormatter(fmt, use_ansi=use_ansi)

    handler = ClickStreamHandler(use_ansi=use_ansi)
    handler.setLevel(INFO)
    handler.setFormatter(formatter)

    logger = DefaultLogger(__name__, level=INFO)
    logger.addHandler(handler)
    return logger


class JobLoggerAdapter(LoggerAdapter):
    def __init__(
            self,
            logger: Logger,
            plain_job_name: str,
            styled_job_name: str,
            use_ansi: bool,
            extra: Union[Dict[str, str], None] = None,
    ):
        if extra is None:
            extra = {}
        updated_extra = {
            "job": styled_job_name if use_ansi else plain_job_name,
            **extra,
        }
        super().__init__(logger=logger, extra=updated_extra)



"""Contains bases of different job types.

You can use the following entities to build your jobs:
    * Class :class:`AsyncJob` or :class:`regta.async_job` decorator
    * Class :class:`ThreadJob` or :class:`regta.thread_job` decorator
    * Class :class:`ProcessJob` or :class:`regta.process_job` decorator
"""

from typing import Awaitable, Callable, Dict, Iterable, Type, Union

from abc import ABC, abstractmethod
import asyncio
from datetime import datetime, timedelta, timezone
from logging import Logger, LoggerAdapter
from multiprocessing import Event as ProcessEventFabric, Process
from multiprocessing.synchronize import Event as ProcessEventObject
from threading import Event as ThreadEvent, Thread

import click
from regta_period import AbstractPeriod

from .enums import JobTypes
from .logging import JobLoggerAdapter, make_default_logger


class AbstractJob(ABC):
    """Interface which every job base implement."""

    interval: Union[timedelta, AbstractPeriod]
    """Interval between every :meth:`.func` call."""
    func: Callable[..., Union[str, None]]
    """The function on which the job will be based. Must be rewritten or passed.
    It'll be called every :attr:`.interval`.
    """

    @abstractmethod
    def execute(self):
        """Execute :attr:`.func` and log result/error.
        It will be called by the job every :attr:`.interval`.
        """
        raise NotImplementedError

    @abstractmethod
    def run(self):
        """Method will be called by scheduler when regta starts. It'll
        call :attr:`.func` every :attr:`.interval`.
        """
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        """Method will be called by the scheduler when regta gets stop
        signal.
        """
        raise NotImplementedError


class BaseJob(AbstractJob):
    """Base job class which implements common logic."""

    interval: Union[timedelta, AbstractPeriod]
    """Interval between every :meth:`.func` call."""
    func: Callable[..., Union[str, None]]
    """The function on which job will be based. Must be rewritten or passed.
    It'll be called every :attr:`.interval`.
    """
    logger: Union[Logger, LoggerAdapter, None] = None
    """Logger writes all results of :attr:`.func` function and
    its exceptions. If the logger isn't specified, regta will use std output.
    """
    args: Iterable = []
    """Args will be passed into :attr:`.func`."""
    kwargs: dict = {}
    """Key-word args will be passed into :attr:`.func`."""

    def __init__(
            self,
            *args,
            logger: Union[Logger, None] = None,
            use_ansi: bool = True,
            **kwargs,
    ):
        self.args = args
        self.kwargs = kwargs

        _logger: Logger = self.logger or logger or make_default_logger(use_ansi=use_ansi)  # type: ignore
        self.logger = JobLoggerAdapter(
            _logger,
            plain_job_name=self.get_plain_job_name(),
            styled_job_name=self.get_styled_job_name(),
            use_ansi=use_ansi,
        )

    def _get_seconds_till_to_execute(self) -> float:
        if isinstance(self.interval, timedelta):
            return self.interval.total_seconds()
        if isinstance(self.interval, AbstractPeriod):
            if self.interval.is_timezone_in_use:
                current_moment = datetime.now(timezone.utc)
            else:
                current_moment = datetime.utcnow()
            return self.interval.get_interval(current_moment).total_seconds()
        raise ValueError(f"Regta doesn't support '{self.interval.__class__.__name__}' type for interval.")

    @abstractmethod
    def execute(self):
        raise NotImplementedError

    @abstractmethod
    def run(self):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError

    @classmethod
    def get_plain_job_name(cls):
        return f"{cls.__module__}:{cls.__name__}"

    @classmethod
    def get_styled_job_name(cls):
        return f"{cls.__module__}:{click.style(cls.__name__, fg='blue')}"


class BaseSyncJob(BaseJob):
    """Class contains common for :class:`ThreadJob` and
    :class:`ProcessJob` blocking.

    .. autoattribute:: interval
    .. autoattribute:: func
    .. autoattribute:: logger
    .. autoattribute:: args
    .. autoattribute:: kwargs
    """

    _blocker_class: Union[Type[ThreadEvent], Callable[..., ProcessEventObject]]

    def __init__(
            self,
            *args,
            logger: Union[Logger, None] = None,
            use_ansi: bool = True,
            **kwargs,
    ):
        super().__init__(*args, logger=logger, use_ansi=use_ansi, **kwargs)
        self._blocker = self._blocker_class()

    def execute(self):
        try:
            res = self.func(*self.args, **self.kwargs)
            self.logger.info(res)
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception(e)

    def __block(self):
        try:
            return self._blocker.wait(self._get_seconds_till_to_execute())
        except KeyboardInterrupt:
            pass
        return True

    def run(self):
        while not self.__block():
            self.execute()

    @abstractmethod
    def stop(self):
        raise NotImplementedError


class ProcessJob(BaseSyncJob, Process):
    """Sync job class based on a process.

    .. autoattribute:: interval
    .. autoattribute:: func
    .. autoattribute:: logger
    .. autoattribute:: args
    .. autoattribute:: kwargs
    .. automethod:: execute
    .. automethod:: run
    """

    _blocker_class = ProcessEventFabric

    def __init__(
            self,
            *args,
            logger: Union[Logger, None] = None,
            use_ansi: bool = True,
            **kwargs,
    ):
        Process.__init__(self)
        super().__init__(*args, logger=logger, use_ansi=use_ansi, **kwargs)

    def stop(self):
        """Stops and terminates job's process. It will be called by the scheduler
        when regta gets a stop signal.
        """
        self._blocker.set()
        self.join()
        self.terminate()


class ThreadJob(BaseSyncJob, Thread):
    """Sync job class based on a thread.

    .. autoattribute:: interval
    .. autoattribute:: func
    .. autoattribute:: logger
    .. autoattribute:: args
    .. autoattribute:: kwargs
    .. automethod:: execute
    .. automethod:: run
    """

    _blocker_class = ThreadEvent

    def __init__(
            self,
            *args,
            logger: Union[Logger, None] = None,
            use_ansi: bool = True,
            **kwargs,
    ):
        Thread.__init__(self)
        super().__init__(*args, logger=logger, use_ansi=use_ansi, **kwargs)

    def stop(self):
        """Stops job's thread. It will be called by the scheduler
        when regta gets a stop signal.
        """
        self._blocker.set()
        self.join()


class AsyncJob(BaseJob):
    """Async job class. Will be executed in an event loop.

    .. autoattribute:: interval
    .. autoattribute:: logger
    .. autoattribute:: args
    .. autoattribute:: kwargs
    """

    func: Callable[..., Awaitable[Union[str, None]]]  # type: ignore
    """The function on which the job will be based. Must be rewritten or passed.
    It'll be called every :attr:`.interval`.
    """

    async def execute(self):
        try:
            res = await self.func(*self.args, **self.kwargs)
            self.logger.info(res)
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception(e)

    async def run(self):
        while True:
            await asyncio.sleep(self._get_seconds_till_to_execute())
            await self.execute()

    async def stop(self): pass


JobHint = Union[AsyncJob, ThreadJob, ProcessJob]


def _make_decorator(_class: Type[JobHint]):
    def decorator(interval: Union[timedelta, AbstractPeriod], *args, **kwargs):
        def wrapper(func: Callable):
            return type(
                func.__name__,
                (_class,),
                {
                    "__module__": func.__module__,
                    "func": staticmethod(func),
                    "interval": interval,
                    "args": args,
                    "kwargs": kwargs,
                }
            )
        return wrapper
    return decorator


async_job = _make_decorator(AsyncJob)
async_job.__doc__ = (
    """Make :class:`AsyncJob` from async function.

    Args:
        interval: Interval between every call.
        *args: Will be passed into the function.
        **kwargs: Will be passed into the function.
    """
)
thread_job = _make_decorator(ThreadJob)
thread_job.__doc__ = (
    """Make :class:`ThreadJob` from function.

    Args:
        interval: Interval between every call.
        *args: Will be passed into the function.
        **kwargs: Will be passed into the function.
    """
)
process_job = _make_decorator(ProcessJob)
process_job.__doc__ = (
    """Make :class:`ProcessJob` from function.

    Args:
        interval: Interval between every call.
        *args: Will be passed into the function.
        **kwargs: Will be passed into the function.
    """
)

jobs_classes: Dict[JobTypes, Type[JobHint]] = {
    JobTypes.ASYNC: AsyncJob,
    JobTypes.THREAD: ThreadJob,
    JobTypes.PROCESS: ProcessJob,
}



from typing import Any, Iterable, List, Sequence, Type, Union

from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
import inspect
from logging import Logger, LoggerAdapter
from pathlib import Path
import re
import string

import click

from .jobs import JobHint, jobs_classes
from .schedulers import Scheduler


def _fix_name(s: str) -> str:
    return s.replace(' ', '_').translate(str.maketrans(string.punctuation, '_'*len(string.punctuation)))


def to_snake_case(s: str) -> str:
    s = _fix_name(s)
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower().replace('__', '_')


def to_camel_case(s: str) -> str:
    s = _fix_name(s)
    return ''.join(word.title() for word in s.split('_'))


def add_init_file(path: Path):
    init_path = path / "__init__.py"
    init_path.touch(exist_ok=True)


def show_jobs_info(
        logger: Union[Logger, LoggerAdapter],
        jobs: Sequence[JobHint] = (),
        classes: Sequence[Type[JobHint]] = (),
        path: Union[Path, None] = None,
        verbose: bool = False,
        use_ansi: bool = True,
):
    jobs_count = (len(jobs)+len(classes))
    end_char = ':' if verbose and jobs_count > 0 else '.'
    count = click.style(
        jobs_count,
        fg='green' if jobs_count > 0 else 'red',
    ) if use_ansi else jobs_count
    path_str = (
        (f" at {click.style(f'{path}/', fg='green')}" if use_ansi else f" at {path}/")
        if path is not None
        else ""
    )
    logger.info(f"[{count}] jobs were found{path_str}{end_char}")

    if verbose:
        for _class in sorted(list(classes) + [job.__class__ for job in jobs], key=lambda _class: _class.__name__):
            logger.info(f"* {_class.get_styled_job_name() if use_ansi else _class.get_plain_job_name()}")


def load_jobs(path: Path) -> List[Type[JobHint]]:
    jobs: List[Type[JobHint]] = []
    for file in path.glob('**/*.py'):
        if file.parts[0][0] == '.':  # skip hidden
            continue

        module_name = ".".join(file.with_suffix("").parts)
        spec = spec_from_file_location(module_name, file)
        if spec is None or spec.loader is None:
            continue
        module = module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore

        jobs.extend(
            cls[1]
            for cls
            in inspect.getmembers(module, inspect.isclass)
            if issubclass(cls[1], tuple(jobs_classes.values()))
            and cls[0] not in (job_class.__name__ for job_class in jobs_classes.values())
        )
    return jobs


def load_object(uri: str) -> Any:
    module_name, object_name = uri.split(':')
    module = import_module(module_name)
    return getattr(module, object_name)


def run_jobs(
        jobs: Iterable[JobHint] = (),
        classes: Iterable[Type[JobHint]] = (),
        logger: Union[Logger, None] = None,
        use_ansi: bool = True,
):
    """Initializes :class:`regta.Scheduler` and starts passed jobs.

    Args:
        jobs: List of job instances.
        classes: List of job classes. This func will make instances from them.
        logger: If a logger isn`t passed, regta will use std output.
        use_ansi: Enable / Disable ANSI colors.

    Raises:
        ValueError: If jobs or classes weren't passed.
    """
    if not jobs and not classes:
        raise ValueError("Jobs or jobs classes missed")

    scheduler = Scheduler()
    for job in jobs:
        job.logger = logger
        scheduler.add_job(job)
    for job_class in classes:
        scheduler.add_job(job_class(logger=logger, use_ansi=use_ansi))
    scheduler.run(block=True)



"""Production-ready scheduler with async, multithreading and multiprocessing support for Python.
See details at the repository on \n https://github.com/SKY-ALIN/regta/
"""

from typing import Callable, Tuple, Type, Union

import asyncio
import logging
from logging import Logger
from pathlib import Path

import click

from . import __version__
from .enums import CodeStyles, JobTypes
from .jobs import AsyncJob, JobHint
from .logging import empty_log_format, JobLoggerAdapter, make_default_logger, prod_log_format
from .templates import generate_decorator_styled_job, generate_oop_styled_job
from .utils import load_jobs, load_object, run_jobs, show_jobs_info

logger_param_help = (
    "Path to logger factory in the following format: "
    "<module>:<logger_factory>. Example: `src.logger:make_jobs_logger`."
)
verbose_param_help = "Set DEBUG level to logger."
no_ansi_param_help = "Disable ANSI colors."
path_param_help = "Path to directory with jobs."
job_type_param_help = "Job type. Defines how the job will use system resources."
code_style_param_help = "Job code style."


def _get_loggers(logger_uri: str, use_ansi: bool, verbose: bool) -> Tuple[Logger, JobLoggerAdapter]:
    logger_factory: Union[Callable[[], Logger], None] = load_object(logger_uri) if logger_uri else None
    logger: Logger = (
        logger_factory()
        if logger_factory
        else make_default_logger(use_ansi=use_ansi, fmt=prod_log_format)
    )

    if verbose:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    wrapped_logger = JobLoggerAdapter(
        logger,
        plain_job_name='regta',
        styled_job_name=click.style('regta', fg='magenta'),
        use_ansi=use_ansi,
    )
    return logger, wrapped_logger


@click.group(help=__doc__)
@click.version_option(__version__)
def main(): pass


@main.command()
@click.option(
    '--path', '-P', 'path',
    default=Path('.'),
    show_default='current directory',
    type=Path,
    help=path_param_help,
)
@click.option(
    '--logger', '-L', 'logger_uri',
    help=logger_param_help,
)
@click.option(
    '--verbose', '-V', 'verbose',
    is_flag=True,
    help=verbose_param_help,
)
@click.option(
    '--no-ansi', 'disable_ansi',
    is_flag=True,
    help=no_ansi_param_help,
)
def run(path: Path, logger_uri: str, disable_ansi: bool, verbose: bool):
    """Start all jobs."""
    use_ansi = not disable_ansi

    logger, wrapped_logger = _get_loggers(logger_uri, use_ansi, verbose)

    classes = load_jobs(path)

    show_jobs_info(classes=classes, verbose=verbose, logger=wrapped_logger, use_ansi=use_ansi)

    try:
        run_jobs(classes=classes, logger=logger, use_ansi=use_ansi)
    except ValueError as e:
        wrapped_logger.info(click.style(str(e), fg='red') if use_ansi else str(e))
    else:
        end_str = "All jobs are closed correctly."
        wrapped_logger.info(click.style(end_str, fg='green') if use_ansi else end_str)


@main.command()
@click.argument('name', type=str)
@click.option(
    '--type', '-T', 'job_type',
    default=JobTypes.THREAD.value,
    show_default=True,
    type=click.Choice([job_type.value for job_type in JobTypes]),
    help=job_type_param_help,
)
@click.option(
    '--style', '-S', 'code_style',
    default=CodeStyles.DECORATOR.value,
    show_default=True,
    type=click.Choice([style.value for style in CodeStyles]),
    help=code_style_param_help,
)
@click.option(
    '--path', '-P', 'path',
    default=Path('jobs/'),
    show_default=True,
    type=Path,
    help=path_param_help,
)
def new(name: str, job_type: str, code_style: str, path: Path):
    """Create a new job by template."""
    style = CodeStyles(code_style)
    if style is CodeStyles.DECORATOR:
        file_name, class_name = generate_decorator_styled_job(name, JobTypes(job_type), path)
    elif style is CodeStyles.OOP:
        file_name, class_name = generate_oop_styled_job(name, JobTypes(job_type), path)
    else:
        raise ValueError("Incorrect job code style")

    click.echo(
        f"{job_type.capitalize()} job {click.style(class_name, fg='blue')} "
        f"in the {code_style} code style have been created at {click.style(path / file_name, fg='green')}."
    )


@main.command(name='list')
@click.option(
    '--path', '-P', 'path',
    default=Path('.'),
    show_default='current directory',
    type=Path,
    help=path_param_help,
)
@click.option(
    '--no-ansi', 'disable_ansi',
    is_flag=True,
    help=no_ansi_param_help,
)
def list_command(path: Path, disable_ansi: bool):
    """Show the list of found jobs."""
    show_jobs_info(
        classes=load_jobs(path),
        path=path,
        verbose=True,
        logger=make_default_logger(use_ansi=not disable_ansi, fmt=empty_log_format),
    )


@main.command()
@click.argument('job_uri')
@click.option(
    '--logger', '-L', 'logger_uri',
    help=logger_param_help,
)
@click.option(
    '--verbose', '-V', 'verbose',
    is_flag=True,
    help=verbose_param_help,
)
@click.option(
    '--no-ansi', 'disable_ansi',
    is_flag=True,
    help=no_ansi_param_help,
)
def execute(job_uri: str, logger_uri: str, verbose: bool, disable_ansi: bool):
    """Execute a single job once and immediately.

    JOB_URI - path to the job in following format: <module>:<job>. Example: `jobs.database_jobs:make_backup`.
    """
    use_ansi = not disable_ansi

    logger, wrapped_logger = _get_loggers(logger_uri, use_ansi, verbose)

    job_class: Type[JobHint] = load_object(job_uri)
    job: JobHint = job_class(logger=logger, use_ansi=use_ansi)

    job_name = job.get_styled_job_name() if use_ansi else job.get_plain_job_name()
    wrapped_logger.warning("Manually initiated execution of %s", job_name)

    if isinstance(job, AsyncJob):
        asyncio.run(job.execute())
    else:
        job.execute()



from typing import Tuple

from pathlib import Path
import random

from jinja2 import Environment, PackageLoader, select_autoescape

from .enums import DecoratorNames, JobTypes
from .jobs import jobs_classes
from .utils import add_init_file, to_camel_case, to_snake_case

env = Environment(
    loader=PackageLoader("regta"),
    autoescape=select_autoescape(),
)


class Templates:
    CLASS_JOB = env.get_template("class_job.template")
    DECORATOR_JOB = env.get_template("decorator_job.template")


def _fix_job_name(name: str) -> str:
    if name[-3:].lower() != 'job':
        name += "-job"
    return name


def _generate_job(name: str, path: Path, template: str) -> str:
    file_name = f"{to_snake_case(name)}.py"

    path.mkdir(parents=True, exist_ok=True)
    add_init_file(path)
    with open(path / file_name, 'w', encoding='utf-8') as job_file:
        job_file.write(template)

    return file_name


def generate_oop_styled_job(name: str, job_type: JobTypes, path: Path) -> Tuple[str, str]:
    name = _fix_job_name(name)

    class_name = to_camel_case(name)
    template = Templates.CLASS_JOB.render(
        job_class=jobs_classes[job_type].__name__,
        class_name=class_name,
        is_async=(job_type == JobTypes.ASYNC),
        seconds=random.randint(2, 60),
    )

    file_name = _generate_job(name, path, template)

    return file_name, class_name


def generate_decorator_styled_job(name: str, job_type: JobTypes, path: Path) -> Tuple[str, str]:
    name = _fix_job_name(name)

    function_name = to_snake_case(name)
    template = Templates.DECORATOR_JOB.render(
        decorator_name=DecoratorNames(job_type).name,
        function_name=function_name,
        is_async=(job_type == JobTypes.ASYNC),
        seconds=random.randint(2, 60),
    )

    file_name = _generate_job(name, path, template)

    return file_name, function_name



"""Contains family of different schedulers.

Use following schedulers to build your system:
    * :class:`AsyncScheduler` for only :class:`regta.AsyncJob`.
    * :class:`SyncScheduler` for only sync jobs (:class:`regta.ThreadJob` or :class:`regta.ProcessJob`).
    * :class:`Scheduler` for all types of jobs (**recommended**).
"""

from typing import List, Union

from abc import ABC, abstractmethod
import asyncio
import signal
from threading import Thread
import time

from .exceptions import IncorrectJobType, StopService
from .jobs import AbstractJob, AsyncJob, ProcessJob, ThreadJob


class AbstractScheduler(ABC):
    """Interface which every scheduler implement."""

    @abstractmethod
    def add_job(self, job: AbstractJob):
        """Add job to scheduler jobs list which will be started.

        Args:
            job: The job will be added.

        Raises:
            IncorrectJobType: If incorrect job type will be passed.
        """
        raise NotImplementedError

    @abstractmethod
    def run(self, block: bool = True):
        """Run scheduler's jobs.

        Args:
            block:
                If True, blocks the current thread. If False, the thread will
                not be blocked and your script will continue.
        """
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        """Stops scheduler's jobs.

        Will be called by :class:`regta.schedulers.SyncBlocking` when regta
        gets stop signal (:exc:`SystemExit`, :exc:`KeyboardInterrupt`, etc.)
        or if the scheduler's thread is not blocked, you can call it yourself.
        """
        raise NotImplementedError


class SyncBlocking(AbstractScheduler, ABC):
    """Contains sync blocking code to add :code:`block` bool var to
    :meth:`.AbstractScheduler.run` to block thread.
    """

    __original_sigterm_handler = None
    __original_sigint_handler = None

    def __stop(self, sig, frame):  # pylint: disable=unused-argument
        self.stop()
        signal.signal(signal.SIGTERM, self.__original_sigterm_handler)
        signal.signal(signal.SIGINT, self.__original_sigint_handler)
        raise StopService

    def _block_main(self):
        self.__original_sigterm_handler = signal.getsignal(signal.SIGTERM)
        signal.signal(signal.SIGTERM, self.__stop)
        self.__original_sigint_handler = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self.__stop)

        while True:
            try:
                time.sleep(1)
            except StopService:
                break


class SyncScheduler(SyncBlocking, AbstractScheduler):
    _thread_jobs: List[ThreadJob] = []
    _process_jobs: List[ProcessJob] = []

    def add_job(self, job: Union[ThreadJob, ProcessJob]):  # type: ignore[override]
        if isinstance(job, ThreadJob):
            self._thread_jobs.append(job)
        elif isinstance(job, ProcessJob):
            self._process_jobs.append(job)
        else:
            raise IncorrectJobType(job, self)

    @staticmethod
    def __start_jobs(jobs: Union[List[ThreadJob], List[ProcessJob]], block: bool):
        for job in jobs:
            job.daemon = not block
            job.start()

    def run(self, block: bool = True):
        self.__start_jobs(self._thread_jobs, block)
        self.__start_jobs(self._process_jobs, block)

        if block:
            self._block_main()

    @staticmethod
    def __stop_jobs(jobs: List[Union[ThreadJob, ProcessJob]]):
        for job in jobs:
            job.stop()

    def stop(self):
        self.__stop_jobs(self._thread_jobs)
        self.__stop_jobs(self._process_jobs)


class AsyncScheduler(AbstractScheduler, Thread):
    _async_jobs: List[AsyncJob] = []
    _async_tasks: List = []

    def __init__(self):
        Thread.__init__(self)
        super().__init__()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def add_job(self, job: AsyncJob):  # type: ignore[override]
        if isinstance(job, AsyncJob):
            self._async_jobs.append(job)
        else:
            raise IncorrectJobType(job, self)

    def run(self, block: bool = True):
        self._async_tasks = [
            self.loop.create_task(job.run())
            for job in self._async_jobs
        ]
        self.loop.run_forever()

    def stop(self):
        """Stop scheduler's jobs."""
        for task in self._async_tasks:
            task.cancel()
        self.loop.call_soon_threadsafe(self.loop.stop)


class Scheduler(SyncBlocking, AbstractScheduler):
    sync_scheduler: Union[SyncScheduler, None] = None
    async_scheduler: Union[AsyncScheduler, None] = None

    def add_job(self, job: Union[AsyncJob, ThreadJob, ProcessJob]):  # type: ignore[override]
        if isinstance(job, AsyncJob):
            if self.async_scheduler is None:
                self.async_scheduler = AsyncScheduler()
            self.async_scheduler.add_job(job)
        elif isinstance(job, (ThreadJob, ProcessJob)):
            if self.sync_scheduler is None:
                self.sync_scheduler = SyncScheduler()
            self.sync_scheduler.add_job(job)
        else:
            raise IncorrectJobType(job, self)

    def run(self, block: bool = True):
        if self.sync_scheduler is not None:
            self.sync_scheduler.run(block=(self.async_scheduler is None and block))
        if self.async_scheduler is not None:
            self.async_scheduler.daemon = not block
            self.async_scheduler.start()
        if block and self.async_scheduler is not None:
            self._block_main()

    def stop(self):
        if self.sync_scheduler is not None:
            self.sync_scheduler.stop()
        if self.async_scheduler is not None:
            self.async_scheduler.stop()



from regta_period import AbstractPeriod, Period, PeriodAggregation, Weekdays

from .jobs import async_job, AsyncJob, process_job, ProcessJob, thread_job, ThreadJob
from .schedulers import Scheduler
from .utils import run_jobs as run

__version__ = '0.3.0'

__all__ = [
    'AbstractPeriod',
    'AsyncJob',
    'Period',
    'PeriodAggregation',
    'ProcessJob',
    'Scheduler',
    'ThreadJob',
    'Weekdays',
    'async_job',
    'process_job',
    'run',
    'thread_job'
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['regta']

package_data = \
{'': ['*'], 'regta': ['templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'asyncio>=3.4.3,<4.0.0',
 'click>=8.1.3,<9.0.0',
 'regta-period>=0.2.0,<0.3.0']

entry_points = \
{'console_scripts': ['regta = regta.console:main']}

setup_kwargs = {
    'name': 'regta',
    'version': '0.3.0',
    'description': 'Production-ready scheduler with async, multithreading and multiprocessing support.',
    'long_description': '# regta\n\n**Production-ready scheduler with async, multithreading and multiprocessing support for Python.**\n\n[![versions](https://img.shields.io/pypi/pyversions/regta.svg)](https://github.com/SKY-ALIN/regta)\n![Code Quality](https://github.com/SKY-ALIN/regta/actions/workflows/code-quality.yml/badge.svg)\n[![PyPI version](https://badge.fury.io/py/regta.svg)](https://pypi.org/project/regta/)\n[![license](https://img.shields.io/github/license/SKY-ALIN/regta.svg)](https://github.com/SKY-ALIN/regta/blob/main/LICENSE)\n\n### Core Features\n\n- **[Various Job Types](https://regta.alinsky.tech/user_guide/make_jobs)** - Create async, thread-based,\n  or process-based jobs depending on your goals.\n\n\n- **[Flexible Intervals](https://regta.alinsky.tech/user_guide/interval_types)** - Use standard `timedelta`\n  or specially designed `Period` for highly responsible jobs.\n\n\n- **[Multi-Paradigm](https://regta.alinsky.tech/user_guide/oop_style)** - Design OOP styled\n  or functional styled jobs.\n\n\n- **[CLI Interface](https://regta.alinsky.tech/cli_reference)** - Regta provides a CLI tool\n  to start, list and create jobs by template.\n\n\n- **[Professional Logging](https://regta.alinsky.tech/user_guide/logging)** - Redefine standard logger\n  and define your own. ANSI coloring is supported.\n\nYou may discover scheduling alternatives and find the comparison with Regta on \n[regta.alinsky.tech/alternatives](https://regta.alinsky.tech/alternatives)\n\n---\n\n### Installation\nInstall using `pip install regta` or `poetry add regta`.\n\nIf you use python < 3.9, then also install backports: `pip install "backports.zoneinfo[tzdata]"`.\n\nYou can check if Regta was installed correctly with the following command `regta --version`.\n\n### Example\n\nTo write async job just use `@regta.async_job()` decorator.\n\n```python\n# jobs/my_jobs.py\n\nfrom datetime import timedelta\nfrom regta import async_job, Period\n\n\n@async_job(Period().every(10).seconds)\nasync def my_period_based_job():\n    return "1. Hello world! This is just a log message."\n\n\n@async_job(timedelta(seconds=10))\nasync def my_timedelta_based_job():\n    return "2. You may use `timedelta` or `Period` as interval."\n\n\n@async_job(Period().on.sunday.at("18:35").by("Asia/Almaty"))\nasync def my_sunday_job():\n    return "3. `Period` is recommended for highly responsible jobs because it does not accumulate shift."\n```\n\nRead more about various job types \n[here](https://regta.alinsky.tech/user_guide/make_jobs).\n\n### Start Up\n\nTo start jobs use `regta run` command:\n\n```shell\n$ regta run\n> [3] jobs were found.\n> 2023-01-08 18:31:00,005 [jobs.my_jobs:my_period_based_job] [INFO] - 1. Hello world! This is just a log message.\n> 2023-01-08 18:31:05,622 [jobs.my_jobs:my_timedelta_based_job] [INFO] - 2. You may use `timedelta` or `Period` as interval.\n.  .  .\n> 2023-01-08 18:34:50,002 [jobs.my_jobs:my_period_based_job] [INFO] - 1. Hello world! This is just a log message.\n> 2023-01-08 18:34:55,689 [jobs.my_jobs:my_timedelta_based_job] [INFO] - 2. You may use `timedelta` or `Period` as interval.\n> 2023-01-08 18:35:00,001 [jobs.my_jobs:my_sunday_job] [INFO] - 3. `Period` is recommended for highly responsible jobs because it does not accumulate shift.\n.  .  .\n```\n\nRead CLI reference [here](https://regta.alinsky.tech/cli_reference).\n\n---\n\nFull documentation and reference are available on \n[regta.alinsky.tech](https://regta.alinsky.tech)\n',
    'author': 'Vladimir Alinsky',
    'author_email': 'Vladimir@Alinsky.tech',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://regta.alinsky.tech',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



from enum import Enum


class JobTypes(Enum):
    ASYNC = 'async'
    THREAD = 'thread'
    PROCESS = 'process'


class DecoratorNames(Enum):
    async_job = JobTypes.ASYNC
    thread_job = JobTypes.THREAD
    process_job = JobTypes.PROCESS


class CodeStyles(Enum):
    OOP = 'oop'
    DECORATOR = 'decorator'
